import logging
import math
from enum import Enum
from threading import Thread

import d3dshot
import pywintypes
import win32api
import win32gui
from ctypes import windll

from fishy.helper.helper import print_exc


class Status(Enum):
    CRASHED = -1
    STOPPED = 0
    RUNNING = 1


class WindowServer:
    """
    Records the game window, and allows to create instance to process it
    """
    Screen = None
    windowOffset = None
    hwnd = None
    status = Status.STOPPED
    d3: d3dshot.D3DShot = d3dshot.create(capture_output="numpy")
    monitor_top_left = None


def init():
    """
    Executed once before the main loop,
    Finds the game window, and calculates the offset to remove the title bar
    """
    # noinspection PyUnresolvedReferences
    try:
        WindowServer.hwnd = win32gui.FindWindow(None, "Elder Scrolls Online")

        monitor_id = windll.user32.MonitorFromWindow(WindowServer.hwnd, 2)
        WindowServer.monitor_top_left = win32api.GetMonitorInfo(monitor_id)["Monitor"][:2]

        rect = win32gui.GetWindowRect(WindowServer.hwnd)
        client_rect = win32gui.GetClientRect(WindowServer.hwnd)
        WindowServer.windowOffset = math.floor(((rect[2] - rect[0]) - client_rect[2]) / 2)
        WindowServer.status = Status.RUNNING

        WindowServer.d3.display = next((m for m in WindowServer.d3.displays if m.hmonitor == monitor_id), None)

    except pywintypes.error:
        logging.error("Game window not found")
        WindowServer.status = Status.CRASHED


def loop():
    """
    Executed in the start of the main loop
    finds the game window location and captures it
    """

    temp_screen = WindowServer.d3.screenshot()

    rect = win32gui.GetWindowRect(WindowServer.hwnd)
    client_rect = win32gui.GetClientRect(WindowServer.hwnd)

    fullscreen = WindowServer.d3.display.resolution[1] == (rect[3] - rect[1])
    title_offset = ((rect[3] - rect[1]) - client_rect[3]) - WindowServer.windowOffset if not fullscreen else 0
    crop = (
        rect[0] + WindowServer.windowOffset - WindowServer.monitor_top_left[0],
        rect[1] + title_offset - WindowServer.monitor_top_left[1],
        rect[2] - WindowServer.windowOffset - WindowServer.monitor_top_left[0],
        rect[3] - WindowServer.windowOffset - WindowServer.monitor_top_left[1]
    )

    WindowServer.Screen = temp_screen[crop[1]:crop[3], crop[0]:crop[2]] if not fullscreen else temp_screen

    if WindowServer.Screen.size == 0:
        logging.error("Don't minimize or drag game window outside the screen")
        WindowServer.status = Status.CRASHED


# noinspection PyBroadException
def run():
    # todo use config
    logging.debug("window server started")
    while WindowServer.status == Status.RUNNING:
        try:
            loop()
        except Exception:
            print_exc()
            WindowServer.status = Status.CRASHED

    if WindowServer.status == Status.CRASHED:
        logging.debug("window server crashed")
    elif WindowServer.status == Status.STOPPED:
        logging.debug("window server stopped")


def start():
    if WindowServer.status == Status.RUNNING:
        return

    init()
    if WindowServer.status == Status.RUNNING:
        Thread(target=run).start()


def screen_ready():
    return WindowServer.Screen is not None or WindowServer.status == Status.CRASHED


def stop():
    WindowServer.status = Status.STOPPED



import ctypes
import logging
import os
import sys

import win32con
import win32gui

import fishy
from fishy.gui import GUI, splash, update_dialog, check_eula
from fishy import helper, web
from fishy.engine.common.event_handler import EngineEventHandler
from fishy.gui.log_config import GuiLogger
from fishy.helper import hotkey
from fishy.helper.active_poll import active
from fishy.helper.config import config
from fishy.helper.hotkey.hotkey_process import hotkey
from fishy.helper.migration import Migration


def check_window_name(title):
    titles = ["Command Prompt", "PowerShell", "Fishy"]
    for t in titles:
        if t in title:
            return True
    return False


# noinspection PyBroadException
def initialize(window_to_hide):
    Migration.migrate()

    helper.create_shortcut_first()

    new_session = web.get_session()

    if new_session is None:
        logging.error("Couldn't create a session, some features might not work")
    logging.debug(f"created session {new_session}")

    try:
        is_admin = os.getuid() == 0
    except AttributeError:
        is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0
    if is_admin:
        logging.info("Running with admin privileges")

    if not config.get("debug", False) and check_window_name(win32gui.GetWindowText(window_to_hide)):
        win32gui.ShowWindow(window_to_hide, win32con.SW_HIDE)
        helper.install_thread_excepthook()
        sys.excepthook = helper.unhandled_exception_logging

    helper.install_required_addons()


def main():
    print("launching please wait...")

    config.init()
    if not check_eula():
        return

    finish_splash = splash.start()
    logger = GuiLogger()
    config.start_backup_scheduler()
    active.init()
    hotkey.init()

    def on_gui_load():
        finish_splash()
        update_dialog.check_update(gui)
        logger.connect(gui)

    window_to_hide = win32gui.GetForegroundWindow()

    bot = EngineEventHandler(lambda: gui)
    gui = GUI(lambda: bot, on_gui_load)

    hotkey.start()

    logging.info(f"Fishybot v{fishy.__version__}")
    initialize(window_to_hide)

    gui.start()
    active.start()

    bot.start_event_handler()   # main thread loop

    hotkey.stop()
    active.stop()
    config.stop()
    bot.stop()


if __name__ == "__main__":
    main()



import logging
import re

import cv2
import numpy as np

detector = cv2.QRCodeDetector()


def image_pre_process(img):
    scale_percent = 100  # percent of original size
    width = int(img.shape[1] * scale_percent / 100)
    height = int(img.shape[0] * scale_percent / 100)
    dim = (width, height)
    img = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    return img


def get_qr_location(image):
    """
    code from https://stackoverflow.com/a/45770227/4512396
    """
    success, points = detector.detect(image)
    if not success:
        return None

    p = points[0]
    # (x, y, x + w, y + h)
    return [int(x) for x in [p[0][0], p[0][1], p[1][0], p[2][1]]]


# noinspection PyBroadException
def get_values_from_image(img):
    h, w = img.shape
    points = np.array([[(0, 0), (w, 0), (w, h), (0, h)]])
    code = detector.decode(img, points)[0]
    return parse_qr_code(code)


# this needs to be updated each time qr code format is changed
def parse_qr_code(code):
    if not code:
        return None
    match = re.match(r'^(\d+\.\d+),(\d+\.\d+),(\d+),(\d+)$', code)
    if not match:
        logging.warning(f"qr code is not what was expected {code}")
        return None
    return [float(match.group(1)), float(match.group(2)), int(match.group(3)), int(match.group(4))]
